#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2014-10-06 17:19:06
# @Author  : Han Zhao (han.zhao@uwaterloo.ca)
# @Link    : https://github.com/KeiraZhao
# @Version : $Id$
import os, sys
import unittest
import theano
import theano.tensor as T
import numpy as np
import scipy as sp
import logging
import time
import csv
import traceback
import random
import argparse

from sklearn.cross_validation import KFold
from scipy.sparse import *

sys.path.append('../source/')

from grcnn import AdaSent
from config import AdaSentConfiger
from wordvec import WordEmbedding
from utils import floatX

charas = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
default_name = ''.join([charas[np.random.randint(0, len(charas))] for _ in xrange(5)])

parser = argparse.ArgumentParser()
parser.add_argument('-s', '--size', help='The size of each batch used to be trained.', 
                    type=int, default=20)
parser.add_argument('-l', '--rate', help='Learning rate of AdaGrad.', 
                    type=float, default=0.02)
parser.add_argument('-e', '--epoch', help='Number of the training epoch.',
                    type=int, default=50)
parser.add_argument('-n', '--name', help='Name used to save the model.', 
                    type=str, default=default_name)
parser.add_argument('-m', '--model', help='Model name to use as initializer.',
                    type=str, default='NONE')
parser.add_argument('-d', '--seed', help='Random seed used for generation.', 
                    type=int, default=42)
parser.add_argument('-k', '--folds', help='K-fold cross-validation to compute the performance score.', 
                    type=int, default=10)
parser.add_argument('config', action='store', type=str)
# Compile and configure all the parameters
args = parser.parse_args()
# Set meta-parameters of the model
np.random.seed(args.seed)
# Logging configuration
# Set the basic configuration of the logging system
log_formatter = logging.Formatter(fmt='%(asctime)s [%(processName)s, %(process)s] [%(levelname)-5.5s]  %(message)s', 
                                  datefmt='%m-%d %H:%M')
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
# File logger
file_handler = logging.FileHandler("{}.log".format(args.name))
file_handler.setFormatter(log_formatter)
file_handler.setLevel(logging.DEBUG)
logger.addHandler(file_handler)
# Stderr logger
std_handler = logging.StreamHandler(sys.stdout)
std_handler.setFormatter(log_formatter)
std_handler.setLevel(logging.DEBUG)
logger.addHandler(std_handler)
# Training and test filename
# This script is used to train and test on the Movie-Review dataset
# Since this data set is not explicitly partitioned into training/test 
# split, we use 10-fold cross-validation to report the final score
mpqa_filename = '../data/MPQA/mpqa.txt'
# Loading and building training and test data set
mpqa_txt, mpqa_label = [], []
start_time = time.time()
# Read all the instances
with file(mpqa_filename, 'r') as fin:
    lines = fin.readlines()
    logger.debug('Number of training instances in the data set: %d' % len(lines))
    for line in lines:
        segments = line.split(',')
        mpqa_txt.append(segments[1])
        mpqa_label.append(int(segments[0]))
# Shuffling all the data
assert len(mpqa_txt) == len(mpqa_label)
data_size = len(mpqa_txt)
random_index = np.arange(data_size)
np.random.shuffle(random_index)
mpqa_txt = list(np.asarray(mpqa_txt)[random_index])
mpqa_label = list(np.asarray(mpqa_label)[random_index])
end_time = time.time()
# Record timing
logger.debug('Time used to load and shuffle MPQA dataset: %f seconds.' % (end_time-start_time))
# Load word-embedding
embedding_filename = '../data/wiki_embeddings.txt'
# Load training/test data sets and wiki-embeddings
word_embedding = WordEmbedding(embedding_filename)
embed_dim = word_embedding.embedding_dim()
start_time = time.time()
blank_index = word_embedding.word2index('</s>')
logger.debug('Blank index: {}'.format(word_embedding.index2word(blank_index)))
# Word-vector representation
mpqa_matrix = []
# Record the sparse input indicator matrix only once for fast computation
mpqa_sparse_select = []
# Embedding for all data instances
for i, sent in enumerate(mpqa_txt):
    words = sent.split()
    words = [word.lower() for word in words]
    vectors = np.zeros((len(words)+2, embed_dim), dtype=np.float32)
    vectors[1:-1] = np.asarray([word_embedding.wordvec(word) for word in words])
    # Record the indices of words in each sentence
    indices = [blank_index]
    indices += [word_embedding.word2index(word) for word in words]
    indices += [blank_index]
    # Construct the sparse one-hot encoding matrix for fast updating word-embedding matrix
    sparse_select = lil_matrix((len(words)+2, word_embedding.dict_size()), dtype=floatX)
    sparse_select[range(len(words)+2), indices] = 1.0
    sparse_select = csc_matrix(sparse_select)
    # Append all the information
    mpqa_matrix.append(vectors)
    mpqa_sparse_select.append(sparse_select)
end_time = time.time()
logger.debug('Time used to build sparse and dense input word-embedding matrices: %f seconds.' % (end_time-start_time))
# Compute the balance of positive/negative count
p_count = np.sum(mpqa_label)
logger.debug('Default positive percentage in dataset: %f' % (float(p_count) / data_size))
logger.debug('Default negative percentage in dataset: %f' % (float(data_size-p_count) / data_size))
# Convert all the data into the form of numpy for easy access
mpqa_label, mpqa_matrix, mpqa_sparse_select = \
np.asarray(mpqa_label), np.asarray(mpqa_matrix), np.asarray(mpqa_sparse_select)
# No explicit partition of training/test data sets, using 10-fold cross-validation
folds = args.folds
cv_partition = KFold(data_size, n_folds=folds)
logger.debug('Partition the whole data set into %d folds.' % args.folds)
logger.debug('Start...')
# Initialize model training configuration
learn_rate = args.rate
batch_size = args.size
epoch = args.epoch
gradient_threshold = 10
embedding_gradient_threshold = 100
# Training using AdaGrad
training_threshold_epoch = epoch / 2
track_training_acc, track_training_cost = [], []
track_test_acc, track_test_cost = [], []
try:
    start_time = time.time()
    sample_size = 0
    fudge_factor = 1e-6
    # accumu matrix for word-embedding matrix
    # hist matrix for word-embedding matrix
    configer = AdaSentConfiger(args.config)
    accumu_embedding = np.zeros((word_embedding.dict_size(), configer.num_input), dtype=floatX)
    hist_embedding = np.zeros((word_embedding.dict_size(), configer.num_input), dtype=floatX)
    # Overall correct count on K-fold cross-validation
    all_train_correct_count, all_test_correct_count = 0, 0
    for cv_id, (train_index, test_index) in enumerate(cv_partition):
        logger.debug('=' * 50)
        # Training
        # Build training and test matrices based on current partition
        train_mpqa_matrix, test_mpqa_matrix = mpqa_matrix[train_index], mpqa_matrix[test_index]
        train_mpqa_label, test_mpqa_label = mpqa_label[train_index], mpqa_label[test_index]
        train_mpqa_sparse_select, test_mpqa_sparse_select = mpqa_sparse_select[train_index], mpqa_sparse_select[test_index]
        train_size, test_size = len(train_index), len(test_index)
        # Output information
        logger.debug('Training on the %d th fold' % cv_id)
        logger.debug('Training size on current partition: %d, test size on current partition: %d' % (train_size, test_size))
        # Building new model on this fold
        # If there is a designated model, using it, else start from scratch
        logger.debug('Building new model on %d th fold.' % cv_id)
        if args.model == 'NONE':
            logger.debug('No designated model, training from scratch...')
            adasent = AdaSent(configer, verbose=True)
        else:
            logger.debug('There is a designated model, loading: {}'.format(args.model))
            adasent = AdaSent.load(args.model)
        # Reload word-embedding matrix
        word_embedding = WordEmbedding(embedding_filename)
        end_time = time.time()
        logger.debug('Time used to building the model: %f seconds.' % (end_time-start_time))
        highest_train_correct_count, highest_test_correct_count = 0, 0
        for i in xrange(epoch):
            # Random shuffle of the training samples at the beginning
            rindex = np.random.permutation(train_size)
            train_mpqa_matrix, train_mpqa_label, train_mpqa_sparse_select = \
            train_mpqa_matrix[rindex], train_mpqa_label[rindex], train_mpqa_sparse_select[rindex]
            # Each epoch includes training on 9 folds and test on 1 fold
            costs = 0.0
            rate = learn_rate
            num_batch = train_size / batch_size
            correct_count = 0
            # Build 
            for k in xrange(num_batch):
                # Clear all the cache        
                accumu_grads = [np.zeros(param.get_value().shape, dtype=np.float32) for param in adasent.params]
                hist_grads = [np.zeros(param.get_value().shape, dtype=np.float32) for param in adasent.params]
                if i > training_threshold_epoch:
                    accumu_embedding[:] = 0.0
                    hist_embedding[:] = 0.0
                for j in xrange(k*batch_size, (k+1)*batch_size):
                    train_sent_rep = train_mpqa_sparse_select[j].dot(word_embedding.embedding)
                    results = adasent.compute_gradient_and_cost(train_sent_rep, train_mpqa_label[j])
                    input_grad = adasent.compute_input_gradient(train_sent_rep, train_mpqa_label[j])
                    grads, cost, pred = results[:-2], results[-2], results[-1]
                    if pred == train_mpqa_label[j]: correct_count += 1
                    costs += cost
                    for accumu_grad, hist_grad, grad in zip(accumu_grads, hist_grads, grads):
                        accumu_grad += grad
                        hist_grad += np.square(grad)
                    ## Update the word-embedding matrix
                    if i > training_threshold_epoch:
                        tmp = train_mpqa_sparse_select[j].T.dot(input_grad)
                        accumu_embedding += tmp
                        hist_embedding += np.square(tmp)
                # Updating model parameters
                for accumu_grad, hist_grad in zip(accumu_grads, hist_grads):
                    accumu_grad /= batch_size
                    accumu_grad /= fudge_factor + np.sqrt(hist_grad)
                    # Gradient clipping
                    accumu_norm = np.sqrt(np.sum(accumu_grad ** 2))
                    if accumu_norm > gradient_threshold:
                        accumu_grad /= accumu_norm / gradient_threshold
                # Updating word-embedding matrix
                if i > training_threshold_epoch:
                    accumu_embedding /= batch_size
                    accumu_embedding /= fudge_factor + np.sqrt(hist_embedding)
                    # Gradient clipping
                    accumu_embedding_norm = np.sqrt(np.sum(accumu_embedding ** 2))
                    if accumu_embedding_norm > embedding_gradient_threshold:
                        accumu_embedding /= accumu_embedding_norm / embedding_gradient_threshold
                    word_embedding._embedding -= rate * accumu_embedding
                adasent.update_params(accumu_grads, rate)
            # Clear all the cache again
            accumu_grads = [np.zeros(param.get_value().shape, dtype=np.float32) for param in adasent.params]
            hist_grads = [np.zeros(param.get_value().shape, dtype=np.float32) for param in adasent.params]
            if i > training_threshold_epoch:
                accumu_embedding[:] = 0.0
                hist_embedding[:] = 0.0
            if num_batch * batch_size < train_size:
                for j in xrange(num_batch*batch_size, train_size):
                    train_sent_rep = train_mpqa_sparse_select[j].dot(word_embedding.embedding)
                    results = adasent.compute_gradient_and_cost(train_sent_rep, train_mpqa_label[j])
                    input_grad = adasent.compute_input_gradient(train_sent_rep, train_mpqa_label[j])
                    grads, cost, pred = results[:-2], results[-2], results[-1]
                    if pred == train_mpqa_label[j]: correct_count += 1
                    costs += cost
                    for accumu_grad, hist_grad, grad in zip(accumu_grads, hist_grads, grads):
                        accumu_grad += grad
                        hist_grad += np.square(grad)
                    ## Update the word-embedding matrix
                    if i > training_threshold_epoch:
                        tmp = train_mpqa_sparse_select[j].T.dot(input_grad)
                        accumu_embedding += tmp
                        hist_embedding += np.square(tmp)
                # Normalizing model parameters
                for accumu_grad, hist_grad in zip(accumu_grads, hist_grads):
                    accumu_grad /= train_size-num_batch*batch_size
                    accumu_grad /= fudge_factor + np.sqrt(hist_grad)
                    # Gradient clipping
                    accumu_norm = np.sqrt(np.sum(accumu_grad ** 2))
                    if accumu_norm > gradient_threshold:
                        accumu_grad /= accumu_norm / gradient_threshold
                # Normalizing and updating word-embedding matrix
                if i > training_threshold_epoch:
                    accumu_embedding /= train_size-num_batch*batch_size
                    accumu_embedding /= fudge_factor + np.sqrt(hist_embedding)
                    # Gradient clipping
                    accumu_embedding_norm = np.sqrt(np.sum(accumu_embedding ** 2))
                    if accumu_embedding_norm > embedding_gradient_threshold:
                        accumu_embedding /= accumu_embedding_norm / embedding_gradient_threshold
                    word_embedding._embedding -= rate * accumu_embedding
                # Updating all the parameters
                adasent.update_params(accumu_grads, rate)
            train_accuracy = float(correct_count) / train_size
            logger.debug('%d CV partition, Training epoch: %d, total cost: %f, Training accuracy = %f' 
                        % (cv_id, i, costs, train_accuracy))
            # Append all the numbers
            track_training_cost.append(costs)
            track_training_acc.append(train_accuracy)
            if correct_count > highest_train_correct_count: highest_train_correct_count = correct_count
            # Testing
            correct_count = 0
            costs = 0.0
            for j in xrange(test_size):
                test_sent_rep = test_mpqa_sparse_select[j].dot(word_embedding.embedding)
                pred = adasent.predict(test_sent_rep)
                cost = adasent.show_cost(test_sent_rep, test_mpqa_label[j])
                if pred == test_mpqa_label[j]: correct_count += 1
                costs += cost
            test_accuracy = float(correct_count) / test_size
            logger.debug('%d CV partition, Training epoch: %d, total cost: %f, Test accuracy: %f' 
                        % (cv_id, i, costs, test_accuracy))
            # Append all the numbers
            track_test_cost.append(costs)
            track_test_acc.append(test_accuracy)
            if correct_count > highest_test_correct_count: highest_test_correct_count = correct_count
            # Check norms of the model parameter
            for param in adasent.params:
                val = param.get_value(borrow=True)
                norm = np.sqrt(np.sum(np.square(val)))
                logger.debug('Parameter: {}, L2-norm: {}'.format(param.name, norm))
            wnorm = np.sqrt(np.sum(np.square(word_embedding._embedding)))
            logger.debug('Parameter: {}, L2-norm: {}'.format('Word-Embedding', wnorm))
        all_train_correct_count += highest_train_correct_count
        all_test_correct_count += highest_test_correct_count
        logger.debug('Highest training accuracy on current fold: %f' % (float(highest_train_correct_count) / train_size))
        logger.debug('Highest test accuracy on current fold: %f' % (float(highest_test_correct_count) / test_size))
except:
    logger.debug('Error appeared!')
    traceback.print_exc(file=sys.stdout)
    logger.debug('-' * 50)
finally:
    logger.debug('After %d-fold cross-validation, highest training accuracy obtained: %f' % (folds, float(all_train_correct_count) / data_size / (folds-1)))
    logger.debug('After %d-fold cross-validation, highest test accuracy obtained: %f' % (folds, float(all_test_correct_count) / data_size))
    logger.debug('Hyperparameter used for this experiment: {}'.format(sys.argv))
    end_time = time.time()
    logger.debug('Time used for training: %f seconds.' % (end_time-start_time))
    AdaSent.save('{}-adasent.model'.format(args.name), adasent)
    # Save all the tracking numbers
    logger.debug('Training and test finished, model saved at: {}-adasent.model'.format(args.name))
    logger.debug('Finished...')
