#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2014-10-06 13:03:51
# @Author  : Han Zhao (han.zhao@uwaterloo.ca)
# @Link    : https://github.com/KeiraZhao
# @Version : $Id$

import os
import numpy as np
import theano
import theano.tensor as T
import time
import cPickle
import logging

import config
import utils
from utils import floatX
from activations import Activation
from mlp import HiddenLayer
from logistic import SoftmaxLayer, LogisticLayer
from score import ScoreLayer

logger = logging.getLogger(__name__)

class GrCNNEncoder(object):
    '''
    (Binary) Gated Recursive Convolutional Neural Network Encoder.
    '''
    def __init__(self, config=None, verbose=True):
        '''
        @config: GRCNNConfiger. Configer used to set the architecture of GRCNNEncoder.
        ''' 
        if verbose: logger.debug('Building Gated Recursive Convolutional Neural Network Encoder...')
        # Scale factor for initializing parameters
        self.scale = config.scale
        # Make theano symbolic tensor for input and model parameters
        self.input = T.matrix(name='GrCNN Encoder input', dtype=floatX)
        # Configure activation function
        self.act = Activation(config.activation)
        fan_in, fan_out = config.num_input, config.num_hidden
        # Initialize model parameters
        # Set seed of the random generator
        np.random.seed(config.random_seed)
        # Projection matrix U
        # Initialize all the matrices using orthogonal matrices        
        U_val = np.random.uniform(low=-1.0, high=1.0, size=(fan_in, fan_out))
        U_val = U_val.astype(floatX)
        U_val *= self.scale
        self.U = theano.shared(value=U_val, name='U', borrow=True)
        self.hidden0 = T.dot(self.input, self.U)

        # W^l, W^r, parameters used to construct the central hidden representation
        Wl_val = np.random.uniform(low=-1.0, high=1.0, size=(fan_out, fan_out))
        Wl_val = Wl_val.astype(floatX)
        Wl_val, _, _ = np.linalg.svd(Wl_val)
        # Wl_val *= self.scale
        self.Wl = theano.shared(value=Wl_val, name='W_l', borrow=True)

        Wr_val = np.random.uniform(low=-1.0, high=1.0, size=(fan_out, fan_out))
        Wr_val = Wr_val.astype(floatX)
        Wr_val, _, _ = np.linalg.svd(Wr_val)
        # Wr_val *= self.scale
        self.Wr = theano.shared(value=Wr_val, name='W_r', borrow=True)
        
        self.Wb = theano.shared(value=np.zeros(fan_out, dtype=floatX), name='Wb', borrow=True)
        
        # G^l, G^r, parameters used to construct the three-way coefficients
        Gl_val = np.random.uniform(low=-1.0, high=1.0, size=(fan_out, 3))
        Gl_val = Gl_val.astype(floatX)
        self.Gl = theano.shared(value=Gl_val, name='G_l', borrow=True)

        Gr_val = np.random.uniform(low=-1.0, high=1.0, size=(fan_out, 3))
        Gr_val = Gr_val.astype(floatX)
        self.Gr = theano.shared(value=Gr_val, name='G_r', borrow=True)

        self.Gb = theano.shared(value=np.zeros(3, dtype=floatX), name='Gb', borrow=True)
        # Save all the parameters into one batch
        self.params = [self.U, self.Wl, self.Wr, self.Wb, self.Gl, self.Gr, self.Gb]
        # Compute the total number of parameters
        self.num_params = reduce(lambda x, y: x+np.prod(y.get_value().shape), self.params, 0)
        # Length of the time sequence
        self.nsteps = self.input.shape[0]
        self.pyramids, _ = theano.scan(fn=self._step_prop, 
                                    sequences=T.arange(self.nsteps-1),
                                    non_sequences=self.nsteps,
                                    outputs_info=[self.hidden0],
                                    n_steps=self.nsteps-1)
        self.output = self.pyramids[-1][0].dimshuffle('x', 0)
        # Compression -- Encoding function
        self.compress = theano.function(inputs=[self.input], outputs=self.output)
        if verbose:
            logger.debug('Finished constructing the structure of {}: '.format(self.__class__.__name__))
            logger.debug('Size of the input dimension: %d' % fan_in)
            logger.debug('Size of the hidden dimension: %d' % fan_out)
            logger.debug('Activation function: %s' % config.activation)

    def _step_prop(self, iter, current_level, nsteps):
        '''
        @current_level: Input matrix at current level. The first dimension corresponds to 
        the timestamp while the second dimension corresponds to the dimension of hidden representation
        '''
        # Build shifted matrix, due to the constraints of Theano.scan, we have to keep the shape of the
        # input and output matrix
        left_current_level = current_level[:nsteps-iter-1]
        right_current_level = current_level[1:nsteps-iter]
        # Compute temporary central hidden representation, of size Txd, but we only care about the first
        # T-1 rows, i.e., we only focus on the (T-1)xd sub-matrix.
        central_current_level = self.act.activate(T.dot(left_current_level, self.Wl) + 
                                                  T.dot(right_current_level, self.Wr) + 
                                                  self.Wb)
        # Compute gating function, of size Tx3. Again, due to the internal limitation of Theano.scan, we cannot
        # reduce the size of the matrix and have to keep the same size, but actually we only want the first (T-1)x3
        # sub-matrix.
        current_gates = T.nnet.softmax(T.dot(left_current_level, self.Gl) + 
                                       T.dot(right_current_level, self.Gr) + 
                                       self.Gb)
        left_gate, central_gate, right_gate = current_gates[:, 0], current_gates[:, 1], current_gates[:, 2]
        # Reshape for broadcasting
        left_gate = left_gate.dimshuffle(0, 'x')
        central_gate = central_gate.dimshuffle(0, 'x')
        right_gate = right_gate.dimshuffle(0, 'x')
        # Build next level of hidden representation using soft combination,
        # matrix of size (T-1)xd
        next_level = left_gate * left_current_level + \
                     right_gate * right_current_level + \
                     central_gate * central_current_level
        return T.set_subtensor(current_level[:nsteps-iter-1], next_level)

    def _step_prop_reduce(self, current_level):
        '''
        @current_level: Input matrix at current level. The first dimension corresponds to 
        the timestamp while the second dimension corresponds to the dimension of hidden representation

        Reduced version of level propagation, much more memory and time efficient implementation, but cannot
        be used inside theano.scan because theano.scan requires that the input and output through timestamps should
        have the same shape.
        '''
        # Build shifted matrix, due to the constraints of Theano.scan, we have to keep the shape of the
        # input and output matrix
        right_current_level = current_level[1:]
        left_current_level = current_level[:-1]
        # Compute temporary central hidden representation, of size Txd, but we only care about the first
        # T-1 rows, i.e., we only focus on the (T-1)xd sub-matrix.
        central_current_level = self.act.activate(T.dot(left_current_level, self.Wl) + 
                                                  T.dot(right_current_level, self.Wr) + 
                                                  self.Wb)
        # Compute gating function, of size Tx3. Again, due to the internal limitation of Theano.scan, we cannot
        # reduce the size of the matrix and have to keep the same size, but actually we only want the first (T-1)x3
        # sub-matrix.
        current_gates = T.nnet.softmax(T.dot(left_current_level, self.Gl) + 
                                       T.dot(right_current_level, self.Gr) + 
                                       self.Gb)
        left_gate, central_gate, right_gate = current_gates[:, 0], current_gates[:, 1], current_gates[:, 2]
        # Reshape for broadcasting
        left_gate = left_gate.dimshuffle(0, 'x')
        central_gate = central_gate.dimshuffle(0, 'x')
        right_gate = right_gate.dimshuffle(0, 'x')
        # Build next level of hidden representation using soft combination,
        # matrix of size (T-1)xd
        next_level = left_gate * left_current_level + \
                     right_gate * right_current_level + \
                     central_gate * central_current_level
        return next_level

    def encode(self, inputM):
        '''
        @input: Theano symbol matrix. Compress the input matrix into output vector.
        '''
        hidden = T.dot(inputM, self.U)
        # Length of the time sequence
        nsteps = inputM.shape[0]
        pyramids, _ = theano.scan(fn=self._step_prop, 
                                    sequences=T.arange(nsteps-1),
                                    non_sequences=nsteps,
                                    outputs_info=[hidden],
                                    n_steps=nsteps-1)
        output = pyramids[-1][0].dimshuffle('x', 0)
        return output

    def L2_loss(self):
        '''
        Return L2 norm of the model parameters.
        '''
        return T.sum(self.U ** 2) + T.sum(self.Wl ** 2) + T.sum(self.Wr ** 2) + \
               T.sum(self.Gl ** 2) + T.sum(self.Gr ** 2)


class GrCNN(object):
    '''
    (Binary) Gated Recursive Convolutional Neural Network Classifier, with GRCNN as the 
    encoder part and MLP as the classifier part.
    '''
    def __init__(self, config=None, verbose=True):
        self.encoder = GrCNNEncoder(config, verbose)
        # Link two parts
        self.params = self.encoder.params
        self.input = self.encoder.input
        self.hidden = self.encoder.output
        # Activation function
        self.act = Activation(config.activation)
        # MLP Component
        self.hidden_layer = HiddenLayer(self.hidden, (config.num_hidden, config.num_mlp), act=Activation(config.hiddenact))
        self.compressed_hidden = self.hidden_layer.output
        # Dropout regularization
        srng = T.shared_randomstreams.RandomStreams(config.random_seed)
        mask = srng.binomial(n=1, p=1-config.dropout, size=self.compressed_hidden.shape)
        self.compressed_hidden *= T.cast(mask, floatX)
        # Accumulate model parameters
        self.params += self.hidden_layer.params
        # Softmax Component
        self.softmax_layer = SoftmaxLayer(self.compressed_hidden, (config.num_mlp, config.num_class))
        self.raw_output = self.softmax_layer.output
        self.pred = self.softmax_layer.pred
        self.params += self.softmax_layer.params
        # Compute the total number of parameters in this model
        self.num_params_encoder = config.num_input * config.num_hidden + \
                                  config.num_hidden * config.num_hidden * 2 + \
                                  config.num_hidden + \
                                  config.num_hidden * 3 * 2 + \
                                  3
        self.num_params_classifier = config.num_hidden * config.num_mlp + \
                                     config.num_mlp + \
                                     config.num_mlp * config.num_class + \
                                     config.num_class
        self.num_params = self.num_params_encoder + self.num_params_classifier
        # Build target function 
        self.truth = T.iscalar(name='label')
        self.learn_rate = T.scalar(name='learning rate')
        # self.cost = self.softmax_layer.NLL_loss(self._truth)
        self.cost = self.softmax_layer.NLL_loss_one(self.truth)
        # Build computational graph and compute the gradient of the target 
        # function with respect to model parameters
        self.gradparams = T.grad(self.cost, self.params)
        # Construct gradient for the input matrix, fine-tuning
        self.input_grads = T.grad(self.cost, self.input)
        # Updates formula for stochastic gradient descent algorithm
        self.updates = []
        for param, gradparam in zip(self.params, self.gradparams):
            self.updates.append((param, param-self.learn_rate*gradparam))
        # Compile theano function
        self.objective = theano.function(inputs=[self.input, self.truth], outputs=self.cost)
        self.predict = theano.function(inputs=[self.input], outputs=self.pred)
        # Compute the gradient of the objective function with respect to the model parameters
        self.compute_gradient_and_cost = theano.function(inputs=[self.input, self.truth], 
                    outputs=self.gradparams+[self.cost, self.pred])
        self.compute_input_gradient = theano.function(inputs=[self.input, self.truth], 
            outputs=self.input_grads)
        # Output function for debugging purpose
        self.show_hidden = theano.function(inputs=[self.input], outputs=self.hidden)
        self.show_compressed_hidden = theano.function(inputs=[self.input], outputs=self.compressed_hidden)
        self.show_output = theano.function(inputs=[self.input], outputs=self.raw_output)
        self.show_cost = theano.function(inputs=[self.input, self.truth], outputs=self.cost)
        if verbose:
            logger.debug('Architecture of GrCNN built finished, summarized as below: ')
            logger.debug('Input dimension: %d' % config.num_input)
            logger.debug('Hidden dimension inside GrCNNEncoder pyramid: %d' % config.num_hidden)
            logger.debug('Hidden dimension of MLP: %d' % config.num_mlp)
            logger.debug('Number of target classes: %d' % config.num_class)
            logger.debug('Number of parameters in encoder part: %d' % self.num_params_encoder)
            logger.debug('Number of parameters in classifier part: %d' % self.num_params_classifier)
            logger.debug('Number of total parameters in this model: %d' % self.num_params)

    def train(self, instance, label):
        '''
        @instance: np.ndarray. Two dimension matrix which corresponds to a time sequence.
        The first dimension along the matrix represents the time dimension while the second 
        dimension along the matrix represents the embedding dimension.
        @label: np.ndarray. 1 dimensional array of int as labels.
        @learn_rate: np.scalar. Learning rate of the stochastic gradient descent algorithm.
        '''
        cost = self.objective(instance, label)
        return cost

    def update_params(self, grads, learn_rate):
        '''
        @grads: [np.ndarray]. List of numpy.ndarray for updating the model parameters.
        @learn_rate: scalar. Learning rate.
        '''
        for param, grad in zip(self.params, grads):
            p = param.get_value(borrow=True)
            param.set_value(p - learn_rate * grad, borrow=True)

    def set_params(self, params):
        '''
        @params: [np.ndarray]. List of numpy.ndarray to set the model parameters.
        '''
        for p, param in zip(self.params, params):
            p.set_value(param, borrow=True)

    @staticmethod
    def save(fname, model):
        '''
        @fname: String. Filename to store the model.
        @model: GrCNN. An instance of GrCNN classifier to be saved.
        '''
        with file(fname, 'wb') as fout:
            cPickle.dump(model, fout)

    @staticmethod
    def load(fname):
        '''
        @fname: String. Filename to load the model.
        '''
        with file(fname, 'rb') as fin:
            model = cPickle.load(fin)
        return model


class GrCNNMatcher(object):
    '''
    (Binary) Gated Recursive Convolutional Neural Network for Matching task, 
    with two GrCNNEncoders as the encoder part and logistic regression as the 
    classifier part.
    '''
    def __init__(self, config=None, verbose=True):
        # Construct two GrCNNEncoders for matching two sentences
        self.encoderL = GrCNNEncoder(config, verbose)
        self.encoderR = GrCNNEncoder(config, verbose)
        # Link two parts
        self.params = []
        self.params += self.encoderL.params
        self.params += self.encoderR.params
        self.inputL = self.encoderL.input
        self.inputR = self.encoderR.input
        # Get output of two GrCNNEncoders
        self.hiddenL = self.encoderL.output
        self.hiddenR = self.encoderR.output
        # Activation function
        self.act = Activation(config.activation)
        # MLP Component
        self.hidden = T.concatenate([self.hiddenL, self.hiddenR], axis=1)
        self.hidden_layer = HiddenLayer(self.hidden, (2*config.num_hidden, config.num_mlp), act=Activation(config.hiddenact))
        self.compressed_hidden = self.hidden_layer.output
        # Accumulate parameters
        self.params += self.hidden_layer.params
        # Dropout parameter
        srng = T.shared_randomstreams.RandomStreams(config.random_seed)
        mask = srng.binomial(n=1, p=1-config.dropout, size=self.compressed_hidden.shape)
        self.compressed_hidden *= T.cast(mask, floatX)
        # Use concatenated vector as input to the logistic regression classifier
        self.logistic_layer = LogisticLayer(self.compressed_hidden, config.num_mlp)
        self.output = self.logistic_layer.output
        self.pred = self.logistic_layer.pred
        # Accumulate parameters
        self.params += self.logistic_layer.params
        # Compute the total number of parameters in this model
        self.num_params_encoder = config.num_input * config.num_hidden + \
                                  config.num_hidden * config.num_hidden * 2 + \
                                  config.num_hidden + \
                                  config.num_hidden * 3 * 2 + \
                                  3
        self.num_params_encoder *= 2
        self.num_params_classifier = 2 * config.num_hidden * config.num_mlp + \
                                     config.num_mlp + \
                                     config.num_mlp + 1
        self.num_params = self.num_params_encoder + self.num_params_classifier
        # Build target function
        self.truth = T.ivector(name='label')
        self.learn_rate = T.scalar(name='learning rate')
        self.cost = self.logistic_layer.NLL_loss(self.truth)
        # Build computational graph and compute the gradient of the target function
        # with respect to model parameters
        self.gradparams = T.grad(self.cost, self.params)
        # Updates formula for stochastic descent algorithm
        self.updates = []
        for param, gradparam in zip(self.params, self.gradparams):
            self.updates.append((param, param-self.learn_rate*gradparam))
        # Compile theano function
        self.objective = theano.function(inputs=[self.inputL, self.inputR, self.truth], outputs=self.cost)
        self.predict = theano.function(inputs=[self.inputL, self.inputR], outputs=self.pred)
        # Compute the gradient of the objective function with respect to the model parameters
        self.compute_cost_and_gradient = theano.function(inputs=[self.inputL, self.inputR, self.truth], 
                                                outputs=self.gradparams+[self.cost, self.pred])
        # Output function for debugging purpose
        self.show_hidden = theano.function(inputs=[self.inputL, self.inputR, self.truth], outputs=self.hidden)
        self.show_compressed_hidden = theano.function(inputs=[self.inputL, self.inputR, self.truth], outputs=self.compressed_hidden)
        self.show_output = theano.function(inputs=[self.inputL, self.inputR, self.truth], outputs=self.output)
        if verbose:
            logger.debug('Architecture of GrCNNMatcher built finished, summarized below: ')
            logger.debug('Input dimension: %d' % config.num_input)
            logger.debug('Hidden dimension inside GrCNNMatcher pyramid: %d' % config.num_hidden)
            logger.debug('Hidden dimension of MLP: %d' % config.num_mlp)            
            logger.debug('Number of parameters in encoder part: %d' % self.num_params_encoder)
            logger.debug('Number of parameters in classifier part: %d' % self.num_params_classifier)
            logger.debug('Number of total parameters in this model: %d' % self.num_params)            

    def update_params(self, grads, learn_rate):
        '''
        @grads: [np.array]. List of numpy.ndarray for updating the model parameters.
        @learn_rate: scalar. Learning rate.
        '''
        for param, grad in zip(self.params, grads):
            p = param.get_value(borrow=True)
            param.set_value(p - learn_rate * grad, borrow=True)

    def set_params(self, params):
        '''
        @params: [np.ndarray]. List of numpy.ndarray to set the model parameters.
        '''
        for p, param in zip(self.params, params):
            p.set_value(param, borrow=True)

    @staticmethod
    def save(fname, model):
        '''
        @fname: String. Filename to store the model.
        @model: GrCNNMatcher. An instance of GrCNNMatcher to be saved.
        '''
        with file(fname, 'wb') as fout:
            cPickle.dump(model, fout)

    @staticmethod
    def load(fname):
        '''
        @fname: String. Filename to load the model.
        '''
        with file(fname, 'rb') as fin:
            model = cPickle.load(fin)
        return model



# AdaSent
## Gating Network implementation
class GatingNetwork(object):
    '''
    Gating-Network component of the AdaSent model. Input of the gating 
    network should be a hierarchy, which is a 2-d matrix with arbitrary
    length on the first dimension and fixed length on the second dimension.
    '''
    def __init__(self, hierarchies, config, verbose=True):
        '''
        Support two kind of gating network:
        1, Softmax classifier, which is a linear weighting function of the hierarchy.
        2, Hidden nonlinear layer + Softmax = MLP, which is a nonlinear weighting 
           function of the hierarchy.
        '''
        if verbose: logger.debug('Creating Gating Network...')
        self.input = hierarchies
        # Configure hidden activation function
        self.act = Activation(config.hiddenact)
        # Parameters of the gating network
        self.params = []
        self.num_params = 0
        # Two kinds
        if config.num_gating > 0:
            self.gating = 'nonlinear'
            self.hidden_layer = HiddenLayer(self.input, (config.num_hidden, config.num_gating),
                                            act=self.act)
            self.hidden = self.hidden_layer.output
            self.weight_layer = theano.shared(name='Weighting vector', 
                                value=np.random.rand(config.num_gating).astype(floatX))
            self.params += self.hidden_layer.params
            self.params += [self.weight_layer]
            self.num_params += self.hidden_layer.num_params
            self.num_params += config.num_gating
        elif config.num_gating == 0:
            self.gating = 'linear'
            self.hidden = self.input
            self.weight_layer = theano.shared(name='Weighting vector', 
                                value=np.random.rand(config.num_hidden).astype(floatX))
            self.params += [self.weight_layer]
            self.num_params += config.num_hidden
        else:
            raise NotImplementedError('{} weighting function is not supported.'.format(config.gating))
        # Softmax weighting function
        self.output = T.nnet.softmax(T.dot(self.hidden, self.weight_layer))

    def L2_loss(self):
        if self.gating == 'nonlinear':
            return self.hidden_layer.L2_loss() + T.sum(self.weight_layer ** 2)
        else: return T.sum(self.weight_layer ** 2)


class AdaSent(object):
    '''
    Self-Adaptive Hierarchical Sentence Model
    '''
    def __init__(self, config, verbose=True):
        '''
        @config: GRCNNConfiger. Configer used to set the architecture of GRCNNEncoder.
        '''
        self.encoder = GrCNNEncoder(config, verbose)
        # Link two parts
        self.input = self.encoder.input
        # Activation function
        self.act = Activation(config.activation)
        # Extract the hierarchical representation, the pyramids, from the encoder
        # Combine the original time series and the compressed time series
        self.pyramids = self.encoder.pyramids
        self.pyramids = T.concatenate([self.encoder.hidden0.dimshuffle('x', 0, 1), 
                                       self.encoder.pyramids])
        self.nsteps = self.pyramids.shape[0]
        # Use another scan function to compress each hierarchical representation
        # into the vector representation
        self.pooling = config.pooling
        self.hierarchies, _ = theano.scan(fn=self._step_compress, 
                                       sequences=[T.arange(self.nsteps, 0, -1), 
                                                  self.pyramids])
        # Global classifier, MLP, mixture of experts
        # Using MLP
        self.hidden_layer = HiddenLayer(self.hierarchies, (config.num_hidden, config.num_mlp), 
                                        act=Activation(config.hiddenact))
        # Adding dropout support
        self.hidden = self.hidden_layer.output
        srng = T.shared_randomstreams.RandomStreams(config.random_seed)
        mask = srng.binomial(n=1, p=1-config.dropout, size=self.hidden.shape)
        self.hidden *= T.cast(mask, floatX)
        # Connect the hidden layer after dropout to a softmax output layer
        self.output_layer = SoftmaxLayer(self.hidden, (config.num_mlp, config.num_class))
        self.experts = self.output_layer.output
        # Gating Network
        # self.weight_layer = theano.shared(name='Weighting vector', value=np.random.rand(config.num_hidden).astype(floatX))
        # self.weights = T.nnet.softmax(T.dot(self.hierarchies, self.weight_layer))
        self.gating_network = GatingNetwork(self.hierarchies, config, verbose)
        self.weights = self.gating_network.output
        # Compute the total number of parameters in the model
        self.num_params = self.encoder.num_params + self.hidden_layer.num_params + \
                          self.output_layer.num_params + self.gating_network.num_params
        # Final decision, convex combination of all the summaries from each layer
        self.score = T.sum(self.experts * self.weights.dimshuffle(1, 0), axis=0)
        # Prediction for classification, since AdaSent processes instance one-by-one
        self.pred = T.argmax(self.score)
        # Stack all the parameters
        self.params = []
        self.params += self.encoder.params
        self.params += self.hidden_layer.params
        self.params += self.output_layer.params
        self.params += self.gating_network.params
        # Build objective function for binary classification problem
        self.truth = T.iscalar(name='label')
        # Cost, Negative Log-Likelihood function, AdaSent works one instance at a time, no need to 
        # add T.mean
        self.cost = -T.log(self.score)[self.truth]
        ## Weight Decay
        if config.weight_decay:
            self.regularizer = self.encoder.L2_loss() + self.hidden_layer.L2_loss() + \
                               self.output_layer.L2_loss() + self.gating_network.L2_loss()
            self.regularizer *= config.weight_decay_parameter
            self.cost += self.regularizer
        # Construct gradient vectors
        self.gradparams = T.grad(self.cost, self.params)
        # Construct gradient for the input matrix, fine-tuning
        self.input_grads = T.grad(self.cost, self.input)
        # Build and compile theano functions
        self.predict = theano.function(inputs=[self.input], outputs=self.pred)
        self.bagging = theano.function(inputs=[self.input], outputs=self.score)
        self.compute_gradient_and_cost = theano.function(inputs=[self.input, self.truth], 
                    outputs=self.gradparams+[self.cost, self.pred])
        self.compute_input_gradient = theano.function(inputs=[self.input, self.truth], 
                    outputs=self.input_grads)
        # Theano functions for debugging purposes
        self.show_weights = theano.function(inputs=[self.input], outputs=self.weights)
        self.show_scores = theano.function(inputs=[self.input], outputs=self.experts)
        self.show_hierarchy = theano.function(inputs=[self.input], outputs=self.hierarchies)
        self.show_prob = theano.function(inputs=[self.input], outputs=self.score)
        self.show_cost = theano.function(inputs=[self.input, self.truth], outputs=self.cost)
        if verbose:
            logger.debug('{} built finished...'.format(self.__class__.__name__))
            logger.debug('Total number of parameters in {}: {}'.format(self.__class__.__name__, self.num_params))

    def _step_compress(self, iter, level):
        '''
        @iter: Int. Index to compress the 0:iter matrices along the first dimension.
        @level: theano symbolic matrix. A time series to be compressed.
        '''
        if self.pooling == 'max':
            return T.max(level[:iter, :], axis=0)
        elif self.pooling == 'averaging':
            return T.mean(level[:iter, :], axis=0)
        else:
            raise NotImplementedError('The pooling method {} is not implemented in {}'.format(self.pooling, AdaSent.__name__))

    def update_params(self, grads, learn_rate):
        '''
        @grads: [np.ndarray]. List of numpy.ndarray for updating the model parameters.
        @learn_rate: scalar. Learning rate.
        '''
        for param, grad in zip(self.params, grads):
            p = param.get_value(borrow=True)
            param.set_value(p - learn_rate * grad, borrow=True)

    def set_params(self, params):
        '''
        @params: [np.ndarray]. List of numpy.ndarray to set the model parameters.
        '''
        for p, param in zip(self.params, params):
            p.set_value(param, borrow=True)

    def deepcopy(self, adasent):
        '''
        @grcnn: AdaSent. Copy the model parameters of another GrCNNMatchScorer and use it.
        '''
        assert len(self.params) == len(adasent.params)
        for p, param in zip(self.params, adasent.params):
            val = param.get_value()
            p.set_value(val)

    @staticmethod
    def save(fname, model):
        '''
        @fname: String. File name to store the model.
        @model: An instance of AdaSent to be saved.
        '''
        with file(fname, 'wb') as fout:
            cPickle.dump(model, fout)

    @staticmethod
    def load(fname):
        '''
        @fname: String. File name to load the model.
        '''
        with file(fname, 'rb') as fin:
            model = cPickle.load(fin)
        return model
